import { EntityRawSchemaArray } from "../raw/Entities";
import { RelationDefinitionRawSchemaArray } from "../raw/RelationDefinitions";
import { RelationDefinitionSchema } from "./RelationDefinitionSchema";
import { Schema } from "./Schema";
import { MatchedRelationDefinition, RelationDefinitionMatcher } from "./Utils";

/**
 * Wrapper around raw entity schemas from the Schema.
 * You should never have to create this object yourself.
 * Instead, retrieve these objects from the methods in the `Schema` class.
 */
export class EntitySchema<
  E extends EntityRawSchemaArray = EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E> = RelationDefinitionRawSchemaArray<E>,
  ERawSchema extends E[number] = E[number]
> {
  /**
   * Wrapper around raw entity schemas from the Schema.
   * You should never have to create this object yourself.
   * Instead, retrieve these objects from the methods in the `Schema` class.
   *
   * @param schema The Schema abstraction class used to create this object
   * @param entityRawSchema The raw entity schema
   */
  constructor(
    private schema: Schema<E, R>,
    private entityRawSchema: ERawSchema
  ) {}

  /**
   * The type of the entity schema. This can also be referred to as the entity schema's name.
   */
  get type(): ERawSchema["type"] {
    return this.entityRawSchema.type;
  }

  /**
   * A short description which can be used as documentation for the entity schema.
   */
  get description(): ERawSchema["description"] {
    return this.entityRawSchema.description;
  }

  /**
   * True if the current entity schema is a primitive type such as string or int
   */
  get final(): ERawSchema["final"] {
    return this.entityRawSchema.final;
  }

  /**
   * Gets the list of relation definitions schemas with this entity as subject.
   * In case a relation definition schema has this entity as both object and subject,
   * there will be dupicates when used along with `objectRelationsDefinitions`.
   */
  get subjectRelationDefinitions(): RelationDefinitionSchema<
    E,
    R,
    MatchedRelationDefinition<
      E,
      R,
      { subject: ERawSchema["type"]; final: false }
    >
  >[] {
    return this.schema.matchRelationDefinitions({
      subject: this.type,
      final: false,
    });
  }

  /**
   * Gets the list of relation definitions schemas with this entity as object.
   * In case a relation definition schema has this entity as both object and subject,
   * there will be dupicates when used along with `subjectRelationDefinitions`.
   */
  get objectRelationsDefinitions(): RelationDefinitionSchema<
    E,
    R,
    MatchedRelationDefinition<
      E,
      R,
      { object: ERawSchema["type"]; final: false }
    >
  >[] {
    return this.schema.matchRelationDefinitions({
      object: this.type,
      final: false,
    });
  }

  /**
   * Finds relations in the Schema matching the partial relation given.
   *
   * @param matcher Partial relation definition to match against
   * @returns Matched relation schema classes
   */
  matchRelationDefinitions<M extends RelationDefinitionMatcher<E>>(
    matcher: M
  ): RelationDefinitionSchema<
    E,
    R,
    MatchedRelationDefinition<E, R, { subject: ERawSchema["type"] }>
  >[] {
    return this.schema.matchRelationDefinitions(matcher);
  }

  /**
   * Gets the list of attribute schemas (relation definition schemas with final at false)
   * for this entity schema.
   */
  get attributes(): RelationDefinitionSchema<
    E,
    R,
    MatchedRelationDefinition<
      E,
      R,
      { subject: ERawSchema["type"]; final: true }
    >
  >[] {
    return this.schema.matchRelationDefinitions({
      final: true,
      subject: this.type,
    });
  }
}
