import { EntityRawSchemaArray } from "../raw/Entities";
import {
  RelationDefinitionRawSchema,
  RelationDefinitionRawSchemaArray,
} from "../raw/RelationDefinitions";
import { RawSchema } from "../raw/Schema";
import {
  EntityRawSchemaByType,
  IsETypeInStrictSchema,
  IsLooseSchema,
} from "../raw/Utils";
import { EntitySchema } from "./EntitySchema";
import { RelationDefinitionSchema } from "./RelationDefinitionSchema";
import { RelationDefinitionMatcher, MatchedRelationDefinition } from "./Utils";

const rawSchemaKey: unique symbol = Symbol("Indexer of the rawSchema");

/**
 * Wrapper around the raw JSON Schema.
 * It is recommended to use this class instead of the raw JSON Schema
 * as it is more robust to schema changes and provides utility functions.
 */
export class Schema<
  E extends EntityRawSchemaArray = EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E> = RelationDefinitionRawSchemaArray<E>
> {
  [rawSchemaKey]: RawSchema<E, R>;

  /**
   * Wrapper around the raw JSON Schema.
   * It is recommended to use this class instead of the raw JSON Schema
   * as it is more robust to schema changes and provides utility functions.
   *
   * @param rawSchema The raw JSON schema to create the wrapper from.
   */
  constructor(rawSchema: RawSchema<E, R>) {
    this[rawSchemaKey] = rawSchema;
  }

  /**
   * Returns the list of entity schemas in the schema using the EntitySchema abstraction class.
   */
  get entities(): EntitySchema<E, R, E[number]>[] {
    return this[rawSchemaKey].entities.map(
      (eSchema) => new EntitySchema(this, eSchema)
    );
  }

  /**
   * Returns the list of relation definitions schemas in the schema
   * using the RelationDefinitionSchema abstraction class.
   */
  get relationsDefinitions(): RelationDefinitionSchema<E, R, R[number]>[] {
    return this[rawSchemaKey].relations_definitions.map(
      (rdefSchema) => new RelationDefinitionSchema(this, rdefSchema)
    );
  }

  /**
   * Returns the entity schema for the given entity type if it exists.
   * When using TypeScript, this method can detect if the given type does not exist
   * in the schema.
   *
   * @param entityType The entity type string
   * @returns A new EntitySchema or null if the type does not exist
   */
  getEntity<EType extends string>(
    entityType: EType
  ): IsLooseSchema<E> extends true
    ? EntitySchema<E, R, E[number]> | null
    : IsETypeInStrictSchema<E, EType> extends true
    ? EntitySchema<E, R, EntityRawSchemaByType<E, EType>>
    : null {
    const entityRawSchema = this[rawSchemaKey].entities.filter(
      (eSchema) => eSchema.type === entityType
    )[0];
    if (!entityRawSchema) {
      // @ts-ignore: Is only null when the given type is not in the schema.
      // Thus the return type will be (EntitySchema | null)
      // but TypeScript does not understand this
      return null;
    }
    // @ts-ignore: see above
    return new EntitySchema<E, R, E[number]>(
      this,
      entityRawSchema as EntityRawSchemaByType<E, EType>
    );
  }

  /**
   * Finds relations definition schemas in the Schema matching the
   * raw partial relation definition schema given.
   *
   * @param matcher Partial raw relation definition schema to match against
   * @returns Array of matched relation definition schema classes
   */
  matchRelationDefinitions<M extends RelationDefinitionMatcher<E>>(
    matcher: M
  ): RelationDefinitionSchema<
    E,
    R,
    MatchedRelationDefinition<E, R, { subject: E[number]["type"] }>
  >[] {
    const matchArray = this[rawSchemaKey].relations_definitions.filter(
      (relDef) => {
        for (const key of Object.keys(
          matcher
        ) as (keyof RelationDefinitionRawSchema<E>)[]) {
          if (matcher[key] !== relDef[key]) {
            return false;
          }
        }
        return true;
      }
    ) as MatchedRelationDefinition<
      E,
      R,
      MatchedRelationDefinition<E, R, { subject: E[number]["type"] }>
    >[];
    return matchArray.map((m) => new RelationDefinitionSchema(this, m));
  }
}
