import {
  EntityRawSchemaArray,
  Client,
  getLooseSchema,
  RelationDefinitionRawSchemaArray,
  Schema,
} from "@cubicweb/client";
import {
  GetListParams,
  GetListResult,
  GetManyParams,
  GetManyReferenceParams,
  GetManyReferenceResult,
  GetManyResult,
  GetOneParams,
  GetOneResult,
  RaRecord,
} from "./types";
import { initVariableKeyAssociation } from "./utils/constants";
import { rowsToRecordList, rowToRecord } from "./utils/dataset";
import {
  getAllRqlParts,
  getFilterRqlParts,
  getSelectionClause,
  getGroupByClause,
  getRestrictionsClause,
  getAllRqlPartsParamsFromMeta,
} from "./utils/getParams";
import { getTotalEntityCount, executeApiRequest } from "./utils/requests";

export const getGet = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  client: Client,
  schema: Schema<E, R>
) => {
  const getList = async <RecordType extends RaRecord>(
    resource: E[number]["type"],
    { pagination, sort, filter }: GetListParams
  ): Promise<GetListResult<RecordType>> => {
    const entitySchema = getLooseSchema(schema).getEntity(resource);
    if (entitySchema == null) {
      return Promise.reject(
        `Entity '${resource}' does not exist in the schema`
      );
    }

    const variableKeyAssociation = initVariableKeyAssociation();
    const {
      selectionTransformers,
      selection,
      groupBy,
      restrictions,
      sortExpression,
      relationsTypes,
    } = getAllRqlParts(entitySchema, sort.field);

    // Handle filters
    const filterRestrictions = getFilterRqlParts(
      filter,
      entitySchema,
      variableKeyAssociation
    );

    restrictions.push(...filterRestrictions);

    const total = await getTotalEntityCount(
      client,
      resource,
      filterRestrictions
    );

    const requestStr = `${getSelectionClause(selection)} ${getGroupByClause(
      groupBy
    )} ORDERBY ${sortExpression} ${sort.order} LIMIT ${
      pagination.perPage
    } OFFSET ${
      (pagination.page - 1) * pagination.perPage
    } ${getRestrictionsClause(restrictions, entitySchema.type)}`;

    return executeApiRequest(client, requestStr).then((rows) => ({
      data: rowsToRecordList(
        rows,
        relationsTypes,
        selectionTransformers
      ) as unknown as RecordType[],
      total,
    }));
  };

  const getOne = async <RecordType extends RaRecord>(
    resource: E[number]["type"],
    params: GetOneParams
  ): Promise<GetOneResult<RecordType>> => {
    const entitySchema = getLooseSchema(schema).getEntity(resource);
    if (entitySchema == null) {
      return Promise.reject(
        `Entity '${resource}' does not exist in the schema`
      );
    }
    const {
      sortData,
      includeSubjectRelations,
      includeObjectRelations,
      knownRelation,
      resolveRelations,
    } = getAllRqlPartsParamsFromMeta(params.meta);
    const {
      selectionTransformers,
      selection,
      groupBy,
      restrictions,
      relationsTypes,
    } = getAllRqlParts(
      entitySchema,
      sortData,
      includeSubjectRelations,
      includeObjectRelations,
      knownRelation,
      resolveRelations
    );

    if (!Number.isInteger(params.id) || params.id < 0) {
      return Promise.reject(
        `'${params.id}' is not an unsigned integer, so not a valid eid`
      );
    }

    const requestStr = `${getSelectionClause(selection)} ${getGroupByClause(
      groupBy
    )} ${getRestrictionsClause(restrictions)}, X eid ${params.id}`;

    return executeApiRequest(client, requestStr).then((rows) => ({
      data: rowToRecord(
        rows[0],
        relationsTypes,
        selectionTransformers
      ) as unknown as RecordType,
    }));
  };

  const getMany = async <RecordType extends RaRecord>(
    resource: E[number]["type"],
    params: GetManyParams
  ): Promise<GetManyResult<RecordType>> => {
    if (params.ids.length === 0) {
      return { data: [] };
    }
    const entitySchema = getLooseSchema(schema).getEntity(resource);
    if (entitySchema == null) {
      return Promise.reject(
        `Entity '${resource}' does not exist in the schema`
      );
    }

    const {
      sortData,
      includeSubjectRelations,
      includeObjectRelations,
      knownRelation,
      resolveRelations,
    } = getAllRqlPartsParamsFromMeta(params.meta);
    const {
      selectionTransformers,
      selection,
      groupBy,
      restrictions,
      relationsTypes,
    } = getAllRqlParts(
      entitySchema,
      sortData,
      includeSubjectRelations,
      includeObjectRelations,
      knownRelation,
      resolveRelations
    );
    const requestStr = `${getSelectionClause(selection)} ${getGroupByClause(
      groupBy
    )} ${getRestrictionsClause(restrictions)}, X eid IN (${params.ids.join(
      ","
    )})`;

    return executeApiRequest(client, requestStr).then((rows) => ({
      data: rowsToRecordList(
        rows,
        relationsTypes,
        selectionTransformers
      ) as unknown as RecordType[],
    }));
  };

  const getManyReference = async <RecordType extends RaRecord>(
    resource: E[number]["type"],
    { target, id, pagination, sort, filter }: GetManyReferenceParams
  ): Promise<GetManyReferenceResult<RecordType>> => {
    const entitySchema = getLooseSchema(schema).getEntity(resource);
    if (entitySchema == null) {
      return Promise.reject(
        `Entity '${resource}' does not exist in the schema`
      );
    }

    const variableKeyAssociation = initVariableKeyAssociation();
    const {
      selectionTransformers,
      selection,
      groupBy,
      restrictions,
      sortExpression,
      relationsTypes,
    } = getAllRqlParts(entitySchema, sort.field, true, true, {
      relation: target,
      value: id,
    });
    // Handle filters
    const filterRestrictions = getFilterRqlParts(
      filter,
      entitySchema,
      variableKeyAssociation
    );
    restrictions.push(...filterRestrictions);

    const total = await getTotalEntityCount(
      client,
      resource,
      filterRestrictions
    );

    const requestStr = `${getSelectionClause(selection)} ${getGroupByClause(
      groupBy
    )} ORDERBY ${sortExpression} ${sort.order} LIMIT ${
      pagination.perPage
    } OFFSET ${
      (pagination.page - 1) * pagination.perPage
    } ${getRestrictionsClause(restrictions)}`;

    return executeApiRequest(client, requestStr).then((rows) => ({
      data: rowsToRecordList(
        rows,
        relationsTypes,
        selectionTransformers
      ) as unknown as RecordType[],
      total,
    }));
  };
  return { getList, getOne, getMany, getManyReference };
};
