import {
  EntityRawSchemaArray,
  RelationDefinitionRawSchemaArray,
  Client,
  Schema,
  getLooseSchema,
  Transaction,
  TransactionQueryRef,
} from "@cubicweb/client";
import {
  RaRecord,
  UpdateManyParams,
  UpdateManyResult,
  UpdateParams,
  UpdateResult,
} from "./types";
import { recursivelyStripEmptyValues } from "./relations";
import { makeUpdateRequest } from "./utils/updateParams";
import { getGet } from "./get";

// FIXME when updating relation, cardinality checks are not possible for the destination. Problems with 1, + and ?:
//    The input will allow any edit because is does not check the cardinality of the destination.
//
// Example: in the museum tutorial schema with the is_in relation, 1* cardinality, with Museum as subject and
//    City as object.
// Editing the city of a museum works as expected:
//    we cannot remove a city from a museum, only change it because the field is required.
// But editing the museums associated with a city does not work as expected (editing is_in from the city object):
//    the input field allows removing a museum from a city because it respects city cardinality constraints (*),
//    but removing a museum from a city will not respect the museum cardinality, the museum will be without a city.
//    We might also add relations from one museum to several cities, which is not allowed.
export const getUpdate = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  client: Client,
  schema: Schema<E, R>
) => {
  const updateTransactionQueries = (
    transaction: Transaction,
    resource: E[number]["type"],
    { data, id, previousData }: UpdateParams
  ): { eidRef: TransactionQueryRef } => {
    const refs = makeUpdateRequest(
      getLooseSchema(schema),
      resource,
      {
        data,
        id,
        previousData: recursivelyStripEmptyValues(previousData),
      },
      transaction
    );
    return { eidRef: refs[refs.length - 1] };
  };

  const update = async <RecordType extends RaRecord>(
    resource: E[number]["type"],
    updateParams: UpdateParams
  ): Promise<UpdateResult<RecordType>> => {
    const transaction = new Transaction();
    const { eidRef: updatedEidRef } = updateTransactionQueries(
      transaction,
      resource,
      updateParams
    );
    const resultSet = await client.executeTransaction(transaction);
    // take the returned value from the last query of the transaction (the update one)
    const updatedEntityRef = resultSet.resolve(updatedEidRef);
    const newData: UpdateResult<RecordType> = await getGet(
      client,
      schema
    ).getOne(resource, {
      id: updatedEntityRef,
    });

    return newData;
  };

  const updateMany = async (
    resource: E[number]["type"],
    { data, ids }: UpdateManyParams
  ): Promise<UpdateManyResult> => {
    console.log("updateMany");
    console.log(resource);
    console.log(data);
    console.log(ids);

    const entitySchema = getLooseSchema(schema).getEntity(resource);
    if (entitySchema == null) {
      return Promise.reject(
        `Entity '${resource}' does not exist in the schema`
      );
    }
    // TODO implement this
    return Promise.reject("updateMany not implemented");
  };

  return { update, updateTransactionQueries, updateMany };
};
