import { Identifier } from "../types";
import { SelectionTransformer } from "./constants";
import { RqlRow, Schema } from "@cubicweb/client";
import isEqual from "lodash/isEqual";
import without from "lodash/without";
import { getRelationFromKey, isIdentifier } from "./relations";

/**
 * Converts a RQL result row to a react-admin record. The id should be the first row item.
 *
 * @param row The rql response row
 * @param relationsTypes The names of relations to use
 * @param selectionTransformers Array of functions used to convert relation values for the UI
 * @returns A react-admin record
 */
export const rowToRecord = (
  row: RqlRow,
  relationsTypes: string[],
  selectionTransformers: SelectionTransformer[]
): { id: Identifier; [other: string]: unknown } => {
  // eid is always the first item in the result

  if (
    row.length !== relationsTypes.length ||
    row.length !== selectionTransformers.length
  )
    throw new Error("Arrays are of different lengths");

  let parsedId = row[0] as Identifier;
  if (typeof parsedId === "string") {
    parsedId = parseInt(parsedId);
    if (isNaN(parsedId)) {
      parsedId = row[0] as Identifier;
    }
  }
  return row.reduce(
    (agg, relationValue, idx) => ({
      [relationsTypes[idx]]: selectionTransformers[idx](relationValue),
      ...agg,
    }),
    { eid: parsedId, id: parsedId }
  );
};

/**
 * Converts a RQL result set to a react-admin record list
 *
 * @param rows The rql response set
 * @param relationsTypes The names of relations to use
 * @param selectionTransformers Array of functions used to convert relations values for the UI
 * @returns A list of react-admin records
 */
export const rowsToRecordList = (
  rows: RqlRow[],
  relationsTypes: string[],
  selectionTransformers: SelectionTransformer[]
) => {
  return rows.map((row) =>
    rowToRecord(row, relationsTypes, selectionTransformers)
  );
};

const _getItemsDiff = (
  item: Identifier[],
  comparedItem: Identifier[]
): Identifier[] => {
  const changedData: Identifier[] = [];
  for (const value of item) {
    if (!comparedItem.includes(value)) {
      changedData.push(value);
    }
  }
  return changedData;
};

/**
 * Finds the changed data between two object with the same keys
 *
 * @param data The new updated data
 * @param previousData The previous data
 * @returns An object containing only the changed data
 */
export const getChangedData = (
  schema: Schema,
  data: Record<string, unknown>,
  previousData: Record<string, unknown>
): {
  changedData: Record<string, unknown>;
  removedData: Record<string, unknown>;
} => {
  const changedData: Record<string, unknown> = {};
  const removedData: Record<string, unknown> = {};

  const isRelation = (key: string) => {
    let relation;
    try {
      relation = getRelationFromKey(key, schema);
    } catch (e) {
      relation = undefined;
    }
    return relation?.final === false;
  };

  // Detect added or modified keys
  for (const key in data) {
    const item = data[key];
    const previousItem = previousData[key];
    if (isEqual(item, previousItem)) {
      continue;
    }

    if (isRelation(key) && item == null) {
      removedData[key] = previousData[key];
    } else if (
      Array.isArray(item) &&
      item.every(isIdentifier) &&
      Array.isArray(previousItem) &&
      previousItem.every(isIdentifier)
    ) {
      // TODO: handle hybrid new, update, removed linked entities here
      changedData[key] = _getItemsDiff(item, previousItem);
      removedData[key] = _getItemsDiff(previousItem, item);
    } else {
      changedData[key] = item;
    }
  }
  // Detect deleted keys
  without(Object.keys(previousData), ...Object.keys(data)).forEach((key) => {
    if (isRelation(key)) {
      removedData[key] = previousData[key];
    } else {
      changedData[key] = undefined;
    }
  });
  return { changedData, removedData };
};
