import {
  Schema,
  Transaction,
  getLooseSchema,
  RQLParams,
  RQLQuery,
  EntitySchema,
  TransactionQueryRef,
} from "@cubicweb/client";
import { Identifier, RaRecord, UpdateParams } from "../types";
import {
  CubicWebReferenceArrayInputValueType,
  getRelationFromKey,
  isIdentifier,
  removeRelationData,
} from "./relations";
import { findAndCreateChildEntities } from "./createParams";
import { getChangedData } from "./dataset";
import { getAttributesNames, getEntitySchemaNames } from "./schema";

type UpdateRelationRequestParamsType = {
  relationsToDelete: [string, string][];
  relationsToSet: [string, string][];
  attributeUpdates: string[];
  attributesValues: RQLParams;
};

type FormData = Partial<RaRecord>;

/**
 * Generates a list of RQL query parts to update the given relation.
 *
 * @param relationName The relation name to get params for
 * @param data The list of ids to update
 * @param resourceRole The role of the resource in the relation
 * @param operationType Specify if the operation should delete or set data
 * @returns
 */
export const generateRqlUpdateRelationsParts = (
  relationName: string,
  data: Identifier[] | undefined,
  resourceRole: "subject" | "object",
  operationType: "delete" | "set"
) => {
  const prefix =
    (resourceRole === "subject" ? "SUB" : "OBJ") +
    (operationType === "set" ? "_SET_REL" : "_DEL_REL");
  const relationsList: Array<[string, string]> = [];
  data?.forEach((i) => {
    if (resourceRole === "subject") {
      relationsList.push([
        `X ${relationName} ${prefix}${relationsList.length}`,
        `${prefix}${relationsList.length} eid ${i}`,
      ]);
    } else {
      relationsList.push([
        `${prefix}${relationsList.length} ${relationName} X`,
        `${prefix}${relationsList.length} eid ${i}`,
      ]);
    }
  });
  return relationsList;
};

/**
 * Creates an object containing ids to delete and ids to add for the given relation.
 *
 * @param relationName The relation name to get params for
 * @param inputData The relation data to update.
 * It can be a single id for reference inputs, or a reference array input object.
 * @param resourceRole The role of the resource in the relation
 * @returns
 */
export const generateRqlUpdateRelationsPartsFromInput = (
  relationName: string,
  inputData: Identifier | CubicWebReferenceArrayInputValueType | undefined,
  resourceRole: "subject" | "object"
) => {
  let setRelationsArray: Identifier[] | undefined;
  if (inputData != null) {
    if (isIdentifier(inputData)) {
      setRelationsArray = [inputData];
    } else {
      setRelationsArray = inputData.selected;
    }
  }
  const deleteRelationsArray =
    inputData != null && !isIdentifier(inputData)
      ? inputData.removed
      : undefined;

  return {
    relationsToDelete: generateRqlUpdateRelationsParts(
      relationName,
      deleteRelationsArray,
      resourceRole,
      "delete"
    ),
    relationsToSet: generateRqlUpdateRelationsParts(
      relationName,
      setRelationsArray,
      resourceRole,
      "set"
    ),
  };
};

/**
 * Generates the RQL query parts to update the entity.
 *
 * @param entitySchema The schema of the entity we are updating
 * @param changedData The data to update
 * @returns An object containing RQL query parts to update attributes and relations
 */
export const generateRqlUpdateParts = (
  entitySchema: EntitySchema,
  changedData: FormData
): UpdateRelationRequestParamsType => {
  const params: UpdateRelationRequestParamsType = {
    relationsToDelete: [],
    relationsToSet: [],
    attributeUpdates: [],
    attributesValues: {},
  };

  getAttributesNames(entitySchema).forEach((relationName) => {
    if (Object.keys(changedData).includes(relationName)) {
      params.attributeUpdates.push(`X ${relationName} %(${relationName})s`);
      const val = changedData[relationName];
      params.attributesValues[relationName] = val == null ? null : val;
    }
  });

  (["subject", "object"] as const).forEach((role) => {
    const relationNamesList = getEntitySchemaNames(
      entitySchema,
      `${role}_relations`
    );
    relationNamesList.forEach((relationName) => {
      if (Object.keys(changedData).includes(relationName)) {
        const relation = getRelationFromKey(relationName, entitySchema);
        const rqlParts = generateRqlUpdateRelationsPartsFromInput(
          relation.type,
          changedData[relationName],
          role
        );
        params.relationsToDelete.push(...rqlParts.relationsToDelete);
        params.relationsToSet.push(...rqlParts.relationsToSet);
      }
    });
  });
  return params;
};

/**
 * Updates an entity using the provided data. It will also recursively create any new child entity.
 *
 * @param schema The CubicWeb schema
 * @param resource The entity type to update
 * @param updateParams The react-admin update parameters
 * @param transaction The transaction used to make the update request.
 * Needed in case we need to create child entities.
 * @returns The pushed queries refs
 */
export const makeUpdateRequest = (
  schema: Schema,
  resource: string,
  { data, id, previousData }: UpdateParams,
  transaction: Transaction
): TransactionQueryRef[] => {
  const entitySchema = getLooseSchema(schema).getEntity(resource);
  if (entitySchema === null) {
    throw new RangeError(`Entity '${resource}' does not exist in the schema`);
  }
  const { changedData, removedData } = getChangedData(
    schema,
    data,
    previousData
  );

  const newChangedData = findAndCreateChildEntities(
    getLooseSchema(schema),
    entitySchema,
    transaction,
    changedData
  );
  const actionOnChangement: {
    [key: string]:
      | Identifier
      | string
      | number
      | { selected: Identifier[]; removed: Identifier[] };
  } = {};
  for (const [key, item] of Object.entries(newChangedData)) {
    if (Array.isArray(item)) {
      if (actionOnChangement[key] === undefined) {
        actionOnChangement[key] = { selected: [], removed: [] };
      }
      (actionOnChangement[key] as { selected: Identifier[] }).selected.push(
        ...item
      );
    } else {
      actionOnChangement[key] = item;
    }
  }
  for (const [key, item] of Object.entries(removedData)) {
    if (actionOnChangement[key] === undefined) {
      actionOnChangement[key] = { selected: [], removed: [] };
    }
    if (Array.isArray(item)) {
      (actionOnChangement[key] as { removed: Identifier[] }).removed.push(
        ...(item as Identifier[])
      );
    } else {
      (actionOnChangement[key] as { removed: Identifier[] }).removed.push(
        item as Identifier
      );
    }
  }
  const {
    relationsToDelete,
    relationsToSet,
    attributeUpdates,
    attributesValues,
  } = generateRqlUpdateParts(entitySchema, actionOnChangement);
  const setAttributesQuery: RQLQuery[] =
    attributeUpdates.length > 0
      ? [
          [
            `SET ${[...attributeUpdates].join(
              ", "
            )} WHERE X is ${resource}, X eid ${id}`,
            attributesValues,
          ],
        ]
      : [];

  const setRelationsQueries: RQLQuery[] = relationsToSet.map(
    ([setPart, wherePart]) => [
      `SET ${setPart} WHERE ${wherePart}, X eid ${id}`,
      {},
    ]
  );

  const deleteRelationsQueries: RQLQuery[] = relationsToDelete.map(
    ([deletePart, wherePart]) => [
      `DELETE ${deletePart} WHERE ${wherePart}, X eid ${id}`,
      {},
    ]
  );

  const queries: RQLQuery[] = [
    ...setAttributesQuery,
    ...setRelationsQueries,
    ...deleteRelationsQueries,
  ];

  const refs: TransactionQueryRef[] = [];
  for (const [rqlQuery, params] of queries) {
    const query = transaction.push(rqlQuery, params);
    refs.push(query.ref());
  }
  removeRelationData(entitySchema, {
    ...data,
    ...newChangedData,
  });
  return refs;
};
